"""
(c) Copyright 2015 Hexoskin
Permission to use, copy, modify, and distribute this software for any purpose with or without fee is hereby granted,
provided that the above copyright notice and this permission notice appear in all copies. The software is provided
"as is" and the author disclaims all warranties with regard to this software including all implied warranties of
merchantability and fitness. In no event shall the author be liable for any special, direct, indirect, or
consequential damages or any damages whatsoever resulting from loss of use, data or profits, whether in an action of
contract, negligence or other tortious action, arising out of or in connection with the use or performance of this
software.

"""

import copy
import json
import os

from load_hx_data.load_hx_data import get_files_specs, load_info, getspec, load_data

VERSION = [0, 12, 0]

DEFAULT_SETTINGS = dict(
    add_quality=False,
    group=False,
    group_raw=False,
    epoch=False,
    header=True,
    skip_raw=False,
    as_nan=False,
    delimiter_tab=False
)

one_hz = ['heart_rate', 'heart_rate_quality', 'breathing_rate', 'breathing_rate_quality', 'minute_ventilation_cl',
          'tidal_volume_ml', 'activity', 'cadence', 'temperature_celcius', 'systolic_pressure', 'SPO2', 'SPO2_quality',
          'BATT_ST_CHANNEL_CHAR', 'TEMP_ST_CHANNEL_CHAR', 'energy_mifflin_keytel']

_FREQ = None


def get_freq():
    return _FREQ


def set_freq(freq):
    global _FREQ
    if freq not in [256, 1000]:
        raise ValueError('Invalid freq')
    _FREQ = freq


set_freq(256)

raw_datatype = ['ECG_I', 'ECG_II', 'ECG_III', 'respiration_thoracic', 'respiration_abdominal', 'acceleration_X',
                'acceleration_Y', 'acceleration_Z', 'PPG', 'ppgdo', 'ppgDisc' 'ppgOOT']

groups = [
    ('group_raw', 'outputDataRaw', raw_datatype, False),
    ('group', 'outputData', one_hz, True)
]


def _print_header(file_specs):
    return ",{} [{}] (/api/datatype/{}/)".format(file_specs['filename'], file_specs['unit'], file_specs['datatype'])


def _convert_file(dirname, dirname_out, settings, file_specs, extension):
    filepath = os.path.join(dirname, file_specs['filename'])
    data, data_quality = load_data_with_quality(filepath, **settings)
    if settings['epoch'] and '_epoch' in file_specs['filename_out']:
        filepath_out = os.path.join(dirname_out, file_specs['filename_out'] + extension)
    elif (dirname != dirname_out or settings['start'] == 0) and not settings['epoch'] \
            and '_epoch' in file_specs['filename_out']:
        filepath_out = os.path.join(dirname_out, file_specs['filename'] + extension)
    elif '_epoch' in file_specs['filename_out']:
        pass
    else:
        filepath_out = os.path.join(dirname_out, file_specs['filename_out'] + extension)
    _save_file(data, data_quality, filepath_out, file_specs, **settings)


def load_data_with_quality(filepath, start=0, add_quality=True, skip_raw=False, **settings):
    file_specs = getspec(filepath)
    data = []
    data_quality = []
    # load_data

    if (skip_raw and file_specs['filename'] in raw_datatype) or \
            (add_quality and os.path.split(filepath)[1].endswith(
                '_quality')):  # if we save quality in heart_rate.csv we don't save heart_rate_quality.csv
        return data, data_quality
    data = load_data(filepath, start)

    # load from file  _quality if necessary
    if add_quality and file_specs['quality_signal_name']:
        filepath = os.path.join(os.path.split(filepath)[0], file_specs['quality_signal_name'])
        data_quality = load_data(filepath, start)
    return data, data_quality


def _save_file(data, data_quality, filepath_out, file_specs, fmt=None, header_string=None, header=True, file_count=0,
               epoch=False, as_nan=False, delimiter_tab=False, **settings):
    delimiter = '\t' if delimiter_tab else ','
    if fmt is None:
        fmt = ['%.8f']
        fmt += ['%.{}f'.format(file_specs['round'])]
    if header:
        if header_string is None:
            header_string = 'time [s]'
            header_string += _print_header(file_specs) if header else ''
    header_string = header_string or ''
    if len(data_quality):
        data = _merge_array(data, data_quality, as_nan)
        filepath_q = os.path.join(os.path.split(filepath_out)[0], file_specs['quality_signal_name'])
        file_specs_quality = getspec(filepath_q)
        header_string += _print_header(file_specs_quality) if header else ''
        fmt += ['%.{}f'.format(0)]
    if file_count == 0:
        mode = 'w'
    else:
        mode = 'a'
        header_string = ''
    if len(data):
        with open(filepath_out, mode) as f:
            f.write(header_string + '\n' if header else '')
            if as_nan:
                fmt = delimiter.join(fmt) + '\n'
                for line in data:
                    f.write(fmt % tuple(line))
            else:
                for line in data:
                    f.write(delimiter.join([fm % a if a is not None else '' for a, fm in zip(line, fmt)]) + '\n')


def _convert_sub_directory(dirname, dirnameout, settings, freq):
    extension = '.txt' if settings.get('delimiter_tab', False) else '.csv'
    if settings['group'] or settings['group_raw']:
        for mode, filepath_out, channels, add_quality in groups:
            if settings[mode]:
                datas = None
                settings['add_quality'] = add_quality
                filepath_out = os.path.join(dirnameout, filepath_out + extension)
                fmt = ['%.8f']
                header = 'time [s]'
                for filename, file_specs in get_files_specs(freq).items():
                    if filename in channels and not \
                            (not settings['add_quality'] and 'quality' in filename):
                        filepath = os.path.join(dirname, file_specs['filename'])
                        data, data_quality = load_data_with_quality(filepath, **settings)
                        if len(data):
                            datas = _merge_array(datas, data, file_specs['dt'], **settings)
                            fmt += ['%.{}f'.format(file_specs['round'])]
                            header += _print_header(file_specs) if settings['header'] else ''
                        if len(data_quality):
                            datas = _merge_array(datas, data_quality, file_specs['dt'], **settings)
                            fmt += ['%.{}f'.format(0)]
                            filepath_q = os.path.join(dirname, file_specs['quality_signal_name'])
                            file_specs_quality = getspec(filepath_q)
                            header += _print_header(file_specs_quality) if settings['header'] else ''
                _save_file(datas, [], filepath_out, file_specs, fmt=fmt, header_string=header, **settings)
    else:
        for filename, file_specs in get_files_specs(freq).items():
            _convert_file(dirname, dirnameout, settings, file_specs, extension)


def _merge_array(in1, in2, dt=None, as_nan=False, **settings):
    if not in1:
        return in2
    in2 = _add_none(in2, dt, as_nan)
    return [l1 + l2[1:] for l1, l2 in zip(in1, in2)]


def _add_none(data, dt, as_nan):
    """allow to put 256, 128,64 Hz in the same file by filling with None"""
    count = {1 / 256: 0, 1 / 128: 1, 1 / 64: 3, 1 / 250: 0, 1 / 125: 1, 1 / 50: 4}
    if data:
        tmp = tuple([float('nan') for a in data[0]]) if as_nan else tuple([None for a in data[0]])
    for a in data:
        yield a
        for i in range(count.get(dt, 0)):
            yield tmp


def convert_directory(directory, settings=None):
    settings = settings or DEFAULT_SETTINGS
    if not os.path.exists(os.path.join(directory, "info.json")):
        raise IOError("{} is missing. Please redownload. Aborted.".format(os.path.join(directory, "info.json")))
    check_version(directory)
    info = load_info(directory)
    freq = info['freq']
    root_start = info.get('start', info.get('start_timestamp', 0))

    dir_list = []
    for subdir in os.listdir(directory):
        subdir = os.path.join(directory, subdir)
        if os.path.isdir(subdir) and os.path.split(subdir)[1].startswith('record_'):
            with open(os.path.join(subdir, "info.json"), "r") as f:
                info_sub = json.load(f)
            dir_list.append([info_sub.get('start', info_sub.get('start_timestamp', 0)), subdir])
    if dir_list:
        dir_list = sorted(dir_list)
        for i, (start_subdir, subpath) in enumerate(dir_list):
            settings2 = copy.copy(settings)
            settings2['header'] = settings2['header'] and i == 0
            settings2['file_count'] = i
            settings2['start'] = start_subdir / freq if settings['epoch'] else (start_subdir - root_start) / freq
            _convert_sub_directory(subpath, directory, settings2, freq)
    else:
        settings['start'] = root_start / freq if settings['epoch'] else 0
        _convert_sub_directory(directory, directory, settings, freq)


def check_version(directory):
    with open(os.path.join(directory, "info.json"), "r") as f:
        info = json.load(f)
    file_version = [int(a) for a in info['decoder_version'].split('.')]
    if VERSION < file_version:
        raise AssertionError("Decoder version {}  too old. redownload decoder at: "
                             "'http://support.hexoskin.com/customer/portal/articles/1758581-how-do-i-convert-binary-"
                             "data-downloaded-from-the-dashboard-'\n".format(VERSION))
